# Copyright 2017-2020 Kamil Sindi
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import inspect
import types
import sys

from pkg_resources import get_distribution, DistributionNotFound


__title__ = 'implements'
__author__ = ('Kamil Sindi <ksindi@ksindi.com>, '
              'Praveen G Shirali <praveengshirali@gmail.com>')
__license__ = 'Apache License, Version 2.0'


try:
    __version__ = get_distribution(__name__).version
except DistributionNotFound:
    # package is not installed
    pass


__all__ = ['Interface', 'implements']


class Interface:
    pass


def implements(interface_cls):
    """Verifies whether the decorated class implements the interface as
    defined by the `interface_cls`.
    """
    def _decorator(cls):
        verify_class_hierarchy(interface_cls, cls)
        errors = []
        errors.extend(verify_methods(interface_cls, cls))
        errors.extend(verify_properties(interface_cls, cls))
        errors.extend(verify_attributes(interface_cls, cls))
        if errors:
            raise NotImplementedError(
                'Found {} errors in implementation:\n- {}\nwith {}'.format(
                    len(errors), '\n- '.join(errors), cls))
        return cls

    return _decorator


def get_mro(cls):
    return cls.mro()[:-1] if cls.mro()[-1] is object else cls.mro()


def verify_class_hierarchy(ifc, cls):
    ifc_mro = get_mro(ifc)
    cls_mro = get_mro(cls)
    common = set(ifc_mro) & set(cls_mro)
    if len(common):
        raise ValueError(
            "Found {} common classes between the implementation and the "
            "interface. Expected none. The implementation class and any "
            "class in its class-hierarchy, must not inherit from the "
            "interface class, or any class from the interface hierarchy. "
            "Common classes: [{}]"
            "".format(len(common), ", ".join([str(s) for s in common]))
        )


def getobj_via_dict(cls, name):
    for c in cls.__mro__:
        if name in c.__dict__:
            return c.__dict__[name]
    return None


def is_classmethod(obj):
    if sys.version_info < (3, 7):
        clsmethod_ident = classmethod
    else:
        clsmethod_ident = (classmethod, types.ClassMethodDescriptorType)
    return isinstance(obj, clsmethod_ident)


def is_staticmethod(obj):
    return isinstance(obj, (staticmethod, types.BuiltinMethodType))


def verify_method_type(method_typer, expected_type,
                       name, ifc_obj, cls_obj, ifc_name, cls_name):
    """Verify a method's type across interface and implementation. Raises
    an exception if they don't match.

    Args:
        method_typer (callable):
            A method type checker (single argument callable) which returns
            True if the supplied argument matches the method type
        expected_type (string):
            A string representation of the expected method type. This is
            used in the exception string
        name (string):
            Name of the attribute being checked
        ifc_obj (object):
            The fetched object from the interface, matched by name
        cls_obj (object):
            The fetched object from the implementation, matched by name
        ifc_name (string):
            Name of the interface class
        cls_name (string):
            Name of the implementation class
    """
    errors = []
    if method_typer(ifc_obj):
        if not method_typer(cls_obj):
            errors.append(
                "'{}' must implement '{}' as {} as defined in interface '{}'"
                "".format(cls_name, name, expected_type, ifc_name)
            )
    return errors


def verify_methods(interface_cls, cls):
    def methods_predicate(m):
        return inspect.isfunction(m) or inspect.ismethod(m)

    errors = []
    for name, method in inspect.getmembers(interface_cls, methods_predicate):
        signature = inspect.signature(method)
        cls_method = getattr(cls, name, None)
        cls_signature = None
        ifc_name = interface_cls.__name__
        cls_name = cls.__name__
        if cls_method and callable(cls_method):
            cls_signature = inspect.signature(cls_method)
        ifc_obj = getobj_via_dict(interface_cls, name)
        cls_obj = getobj_via_dict(cls, name)

        method_types_to_check = [
            (is_classmethod, "a classmethod"),
            (is_staticmethod, "a staticmethod"),
            (inspect.isasyncgenfunction, "an async genenerator-function"),
            (inspect.isgeneratorfunction, "a generator-function"),
            (inspect.iscoroutinefunction, "a coroutine-function")
        ]
        for (method_typer, expected_type) in method_types_to_check:
            errors.extend(
                verify_method_type(method_typer, expected_type,
                                   name, ifc_obj, cls_obj, ifc_name, cls_name)
            )

        if cls_signature != signature:
            errors.append(
                "'{}' must implement method '{}{}' defined in interface '{}'"
                .format(cls_name, name, signature, ifc_name)
            )
    return errors


def verify_properties(interface_cls, cls):
    errors = []
    prop_attrs = dict(fget='getter', fset='setter', fdel='deleter')
    descriptors = inspect.getmembers(interface_cls, inspect.isdatadescriptor)
    for name, prop in descriptors:
        cls_prop = getattr(cls, name, None)
        for attr in prop_attrs:
            # instanceof doesn't work for class function comparison
            ifc_prop_obj = getattr(prop, attr, None)
            cls_prop_obj = getattr(cls_prop, attr, None)
            if ifc_prop_obj:
                cls_name = cls.__name__
                ifc_name = interface_cls.__name__
                proptype = prop_attrs[attr]

                # -- verify presence and type of data-descriptors
                if type(ifc_prop_obj) != type(cls_prop_obj):
                    errors.append(
                        "'{}' must implement a {} for property '{}' defined "
                        "in interface '{}'"
                        "".format(cls_name, proptype, name, ifc_name)
                    )
                    continue

                # -- verify signatures of data-descriptors
                ifc_prop_sig = inspect.signature(ifc_prop_obj)
                cls_prop_sig = None
                if callable(cls_prop_obj):
                    cls_prop_sig = inspect.signature(cls_prop_obj)
                if ifc_prop_sig != cls_prop_sig:
                    errors.append(
                        "'{}' must implement a {} for property '{}' with the "
                        "same signature as defined in interface '{}'"
                        "".format(cls_name, proptype, name, ifc_name)
                    )
    return errors


def verify_attributes(interface_cls, cls):
    errors = []
    interface_attributes = get_attributes(interface_cls)
    cls_attributes = get_attributes(cls)
    for missing_attr in interface_attributes - cls_attributes:
        errors.append(
            "'{}' must have class attribute '{}' defined in interface '{}'"
            .format(cls.__name__, missing_attr, interface_cls.__name__)
        )
    return errors


def get_attributes(cls):
    boring = dir(type('dummy', (object,), {}))
    return set(item[0] for item in inspect.getmembers(cls)  # skipcq: PTC-W0015
               if item[0] not in boring and not callable(item[1]))



# Copyright 2017-2020 Kamil Sindi
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import sys
import pytest

from implements import Interface, implements, get_mro


py36 = pytest.mark.skipif(sys.version_info < (3, 6), reason='requires py3.6')


def test_empty():
    class FooInterface(Interface):
        pass

    @implements(FooInterface)
    class FooImplementation:
        pass


def test_with_args_kwargs():
    class FooInterface(Interface):
        def foo(self, a, *args, b=1, **kwargs):
            pass

    with pytest.raises(NotImplementedError):
        @implements(FooInterface)
        class FooImplementationFail:
            def foo(self, a, *args, b=7):
                pass

    @implements(FooInterface)
    class FooImplementationPass:
        def foo(self, a, *args, b=1, **kwargs):
            pass


def test_with_kwarg_only():
    class FooInterface(Interface):
        def foo(self, a, *, b):
            pass

    with pytest.raises(NotImplementedError):
        @implements(FooInterface)
        class FooImplementationFail:
            def foo(self, a, b):
                pass

    @implements(FooInterface)
    class FooImplementationPass:
        def foo(self, a, *, b):
            pass


def test_property():
    class FooInterface(Interface):
        @property
        def foo(self):
            pass

    with pytest.raises(NotImplementedError):
        @implements(FooInterface)
        class FooImplementationFail:
            def foo(self):
                pass

    @implements(FooInterface)
    class FooImplementationPass:
        @property
        def foo(self):
            pass


def test_property_inverse():
    class FooInterface(Interface):
        def foo(self):
            pass

    with pytest.raises(NotImplementedError):
        @implements(FooInterface)
        class FooImplementationFail:
            @property
            def foo(self):
                pass


def test_setters():
    class FooInterface(Interface):
        @property
        def foo(self):
            pass

        @foo.setter
        def foo(self, val):
            pass

    with pytest.raises(NotImplementedError):
        @implements(FooInterface)
        class FooImplementationFail:
            @property
            def foo(self):
                pass

    @implements(FooInterface)
    class FooImplementationPass:
        @property
        def foo(self):
            pass

        @foo.setter
        def foo(self, val):
            pass


def test_deleters():
    class FooInterface(Interface):
        @property
        def foo(self):
            pass

        @foo.deleter
        def foo(self, val):
            pass

    with pytest.raises(NotImplementedError):
        @implements(FooInterface)
        class FooImplementationFail:
            @property
            def foo(self):
                pass

    @implements(FooInterface)
    class FooImplementationPass:
        @property
        def foo(self):
            pass

        @foo.deleter
        def foo(self, val):
            pass


def test_implementation_implements_more_descriptors():
    class FooInterface(Interface):
        @property
        def foo(self):
            pass

    #   An implementation must implement all data descriptors defined in
    #   the interface, however, the implementation could define more.
    #
    #   The case below must not generate errors because FooImplementationPass
    #   defines a foo.setter which isn't defined by FooInterface
    @implements(FooInterface)
    class FooImplementationPass:
        @property
        def foo(self):
            pass

        @foo.setter
        def foo(self, val):
            pass


def test_missing_method():
    class FooInterface(Interface):
        def foo(self):
            pass

    with pytest.raises(NotImplementedError):
        @implements(FooInterface)
        class FooImplementationFail:
            pass

    @implements(FooInterface)
    class FooImplementationPass:
        def foo(self):
            pass


def test_missing_argument():
    class FooInterface(Interface):
        def foo(self, arg):
            pass

    with pytest.raises(NotImplementedError):
        @implements(FooInterface)
        class FooImplementationFail:
            def foo(self):
                pass

    @implements(FooInterface)
    class FooImplementationPass:
        def foo(self, arg):
            pass


def test_renamed_argument():
    class FooInterface(Interface):
        def foo(self, arg):
            pass

    with pytest.raises(NotImplementedError):
        @implements(FooInterface)
        class FooImplementationFail:
            def foo(self, arrrrg):
                pass

    @implements(FooInterface)
    class FooImplementationPass:
        def foo(self, arg):
            pass


def test_extra_argument():
    class FooInterface(Interface):
        def foo(self, arg):
            pass

    with pytest.raises(NotImplementedError):
        @implements(FooInterface)
        class FooImplementationFail:
            def foo(self, arg, ument):
                pass

    @implements(FooInterface)
    class FooImplementationPass:
        def foo(self, arg):
            pass


def test_different_defaults():
    class FooInterface(Interface):
        def foo(self, arg=7):
            pass

    with pytest.raises(NotImplementedError):
        @implements(FooInterface)
        class FooImplementationFail:
            def foo(self, arg=8):
                pass

    @implements(FooInterface)
    class FooImplementationPass:
        def foo(self, arg=7):
            pass


def test_different_order():
    class FooInterface(Interface):
        def foo(self, a, b):
            pass

    with pytest.raises(NotImplementedError):
        @implements(FooInterface)
        class FooImplementationFail:
            def foo(self, b, a):
                pass

    @implements(FooInterface)
    class FooImplementationPass:
        def foo(self, a, b):
            pass


def test_missing_kwargs():
    class FooInterface(Interface):
        def foo(self, **kwargs):
            pass

    with pytest.raises(NotImplementedError):
        @implements(FooInterface)
        class FooImplementationFail:
            def foo(self):
                pass

    @implements(FooInterface)
    class FooImplementationPass:
        def foo(self, **kwargs):
            pass


def test_missing_property():
    class FooInterface(Interface):
        @property
        def foo(self):
            pass

    with pytest.raises(NotImplementedError):    # missing method
        @implements(FooInterface)
        class FooImplementationFail1:           # skipcq: PYL-W0612
            pass

    with pytest.raises(NotImplementedError):    # missing property decorator
        @implements(FooInterface)
        class FooImplementationFail2:           # skipcq: PYL-W0612
            def foo(self):
                pass

    @implements(FooInterface)
    class FooImplementationPass:
        @property
        def foo(self):
            pass


def test_bad_constructor():
    class FooInterface(Interface):
        def __init__(self, a):
            pass

    with pytest.raises(NotImplementedError):
        @implements(FooInterface)
        class FooImplementationFail:
            def __init__(self):
                pass

    @implements(FooInterface)
    class FooImplementationPass:
        def __init__(self, a):
            pass


def test_multiple_errors():
    class FooInterface(Interface):
        @property
        def foo(self):
            pass

        def __init__(self, a):
            pass

    # Bad constructor, missing method getter, and missing class attribute (3)
    match = r'^Found 3 errors in implementation:\n- .+\n- .+\n- .+\nwith .+'
    with pytest.raises(NotImplementedError, match=match):
        @implements(FooInterface)
        class FooImplementationFail:           # skipcq: PYL-W0612
            def __init__(self):
                pass


def test_static():
    class FooInterface(Interface):
        @staticmethod
        def foo(a, b, c):
            pass

    with pytest.raises(NotImplementedError):
        @implements(FooInterface)
        class FooImplementationFail1:           # skipcq: PYL-W0612
            pass                    # missing foo

    with pytest.raises(NotImplementedError):
        @implements(FooInterface)
        class FooImplementationFail2:           # skipcq: PYL-W0612
            # skipcq: PYL-E0213
            def foo(a, b, c):       # missing staticmethod decorator
                pass

    with pytest.raises(NotImplementedError):
        @implements(FooInterface)
        class FooImplementationFail3:           # skipcq: PYL-W0612
            @classmethod            # classmethod instead of staticmethod
            def foo(cls, a, b, c):  # decorator-check fails before signature
                pass

    with pytest.raises(NotImplementedError):
        @implements(FooInterface)
        class FooImplementationFail4:           # skipcq: PYL-W0612
            @staticmethod
            def foo(m, n, o):       # staticmethod, but wrong signature
                pass

    @implements(FooInterface)
    class FooImplementationPass:
        @staticmethod
        def foo(a, b, c):
            pass


def test_classmethods():
    class FooInterface(Interface):
        @classmethod
        def foo(cls, a, b, c):
            pass

    with pytest.raises(NotImplementedError):
        @implements(FooInterface)
        class FooImplementationFail1:           # skipcq: PYL-W0612
            pass                    # missing foo

    with pytest.raises(NotImplementedError):
        @implements(FooInterface)
        class FooImplementationFail2:           # skipcq: PYL-W0612
            # skipcq: PYL-E0213
            def foo(cls, a, b, c):  # missing classmethod decorator
                pass

    with pytest.raises(NotImplementedError):
        @implements(FooInterface)
        class FooImplementationFail3:           # skipcq: PYL-W0612
            @staticmethod           # staticmethod instead of classmethod
            def foo(a, b, c):       # decorator-check fails before signature
                pass

    with pytest.raises(NotImplementedError):
        @implements(FooInterface)
        class FooImplementationFail4:           # skipcq: PYL-W0612
            @classmethod
            def foo(cls, m, n, o):   # classmethod, but wrong signature
                pass

    @implements(FooInterface)
    class FooImplementationPass:
        @classmethod
        def foo(cls, a, b, c):
            pass


def test_classmethod_signature_match():
    # For a classmethod, inspect.signature returns a signature with the first
    # element (cls) stripped. A classmethod with signature (cls, a, b, c) has
    # signature equivalence with a regular method with signature (a, b, c)
    #
    # Example:
    from inspect import signature

    class TestA:
        @classmethod
        def foo(cls, a, b, c):
            pass

    class TestB:
        # skipcq: PYL-E0213
        def foo(a, b, c):
            pass

    assert signature(TestA.foo) == signature(TestB.foo)

    # The test below ensures that the above case is flagged
    class FooInterface(Interface):
        @classmethod
        def foo(cls, a, b, c):
            pass

    with pytest.raises(NotImplementedError):
        @implements(FooInterface)
        class FooImplementationFail:
            # skipcq: PYL-E0213
            def foo(a, b, c):
                pass


def test_staticmethod_classmethod_with_decorator():
    class FooBarInterface(Interface):
        @staticmethod
        def foo(a, b, c):
            pass

        @classmethod
        def bar(cls, a, b, c):
            pass

    import functools

    def decorator(func):
        @functools.wraps(func)
        def inner(*args, **kwargs):
            return func(*args, **kwargs)
        return inner

    @implements(FooBarInterface)
    class FooBarImplementationPass:
        @staticmethod
        @decorator
        def foo(a, b, c):
            pass

        @classmethod
        @decorator
        def bar(cls, a, b, c):
            pass


def test_kwargs_only():
    class FooInterface(Interface):
        def foo(self, *, a):
            pass

    with pytest.raises(NotImplementedError):
        @implements(FooInterface)
        class FooImplementation:
            def foo(self, a):
                pass


def test_multiple_interfaces():
    class FooInterface(Interface):
        def foo(self):
            pass

    class BarInterface(Interface):
        def bar(self):
            pass

    with pytest.raises(NotImplementedError):
        @implements(BarInterface)
        @implements(FooInterface)
        class FooImplementationNoBar:
            def foo(self, a):
                pass

    with pytest.raises(NotImplementedError):
        @implements(BarInterface)
        @implements(FooInterface)
        class FooImplementationNoFoo:
            def bar(self, a):
                pass

    @implements(BarInterface)
    @implements(FooInterface)
    class FooImplementation:
        def foo(self):
            pass

        def bar(self):
            pass


def test_interface_name_collision():
    class Foo1Interface(Interface):
        def foo(self):
            pass

    class Foo2Interface(Interface):
        def foo(self):
            pass

    @implements(Foo2Interface)
    @implements(Foo1Interface)
    class FooImplementation:
        def foo(self):
            pass


def test_interface_name_and_signature_collision():
    class Foo1Interface(Interface):
        def foo(self):
            pass

    class Foo2Interface(Interface):
        def foo(self) -> str:
            return 'foo'

    # Two interfaces with different signatures for a given method will
    # always result in failure for the implementing class, as the
    # implemented method's signature can only satisfy one of the interfaces.

    with pytest.raises(NotImplementedError):
        @implements(Foo2Interface)
        @implements(Foo1Interface)
        class FooImplementationFail:
            def foo(self):
                pass


def test_interface_inheritance():
    class BaseInterface(Interface):
        def bar(self):
            pass

    class FooInterface(BaseInterface):
        def foo(self):
            pass

    with pytest.raises(NotImplementedError):
        @implements(FooInterface)
        class FooImplementationFail:
            def foo(self):
                pass

    @implements(FooInterface)
    class FooImplementationPass:
        def foo(self):
            pass

        def bar(self):
            pass


def test_class_inheritance():
    class FooInterface(Interface):
        def foo(self):
            pass

    @implements(FooInterface)
    class ParentImplementation:
        def foo(self):
            pass

    @implements(FooInterface)
    class ChildImplementation(ParentImplementation):
        pass


def test_class_multiple_inheritance():
    # --------- INTERFACES -----------------------------------------------
    #
    class FooInterface(Interface):
        def foo(self, final):
            pass

    class BarInterface(Interface):
        def bar(self, final):
            pass

    class FooBarInterface(FooInterface, BarInterface):
        pass

    # --------- IMPLEMENTATION -------------------------------------------
    #
    class BaseFooImplementation:        # must get overridden
        def foo(self, override, my, args):
            pass

    @implements(FooInterface)
    class FooImplementation(BaseFooImplementation):
        def foo(self, final):           # skipcq: PYL-W0221
            pass

    @implements(BarInterface)
    class BarImplementation:
        def bar(self, final):
            pass

    with pytest.raises(NotImplementedError):
        @implements(FooBarInterface)
        class SubFooImplementation(FooImplementation):  # foo, no bar
            pass

    @implements(FooInterface)
    @implements(BarInterface)
    @implements(FooBarInterface)
    class FooBarImplementation(FooImplementation, BarImplementation):
        pass


def test_rtn_type_annotation():
    class FooInterface(Interface):
        def foo(self) -> str:
            pass

    with pytest.raises(NotImplementedError):
        @implements(FooInterface)
        class FooImplementationFail:
            def foo(self) -> int:
                pass

    @implements(FooInterface)
    class FooImplementationPass:
        def foo(self) -> str:
            pass


def test_arg_type_annotation():
    class FooInterface(Interface):
        def foo(self, arg: str):
            pass

    with pytest.raises(NotImplementedError):
        @implements(FooInterface)
        class FooImplementationFail:
            def foo(self, arg: int):
                pass

    @implements(FooInterface)
    class FooImplementationPass:
        def foo(self, arg: str):
            pass


def test_other_decorator_compat():
    def decorator(cls):
        class Wrapper:
            def __init__(self, *args):
                self.wrapped = cls(*args)

            def __getattr__(self, name):
                print('Getting the {} of {}'.format(name, self.wrapped))
                return getattr(self.wrapped, name, None)

        return Wrapper

    class FooInterface(Interface):
        def foo(self):
            pass

    with pytest.raises(NotImplementedError):
        @implements(FooInterface)
        @decorator
        class FooImplementationFail:
            def __init__(self, x, y):
                self.x = x
                self.y = y

            def foo(self):
                pass

    @decorator
    @implements(FooInterface)
    class FooImplementationPass:
        def __init__(self, x, y):
            self.x = x
            self.y = y

        def foo(self):
            pass


def test_magic_methods():
    class FooInterface(Interface):
        def __add__(self, other):
            pass

    with pytest.raises(NotImplementedError):
        @implements(FooInterface)
        class FooImplementationFail:
            pass

    @implements(FooInterface)
    class FooImplementationPass:
        def __add__(self, other):
            pass


def test_attributes():
    class FooInterface(Interface):
        a = None

    with pytest.raises(NotImplementedError):
        @implements(FooInterface)
        class FooImplementationFail:
            pass

    @implements(FooInterface)
    class FooImplementationPass:
        a = 1
        b = 2


def test_async():
    class AsyncInterface:
        async def __aenter__(self):
            return self

        async def __aexit__(self, *args, **kwargs):
            pass

    with pytest.raises(NotImplementedError):
        @implements(AsyncInterface)
        class AsyncImplementation:
            pass


def test_async_method():
    class AsyncFooInterface:
        async def foo(self):
            pass

    with pytest.raises(NotImplementedError):
        @implements(AsyncFooInterface)
        class FooImplementationFail:                # skipcq: PYL-W0612
            def foo(self):
                pass

    @implements(AsyncFooInterface)
    class AsyncFooImplementation:                   # skipcq: PYL-W0612
        async def foo(self):
            pass


def test_generator():
    class GenFooInterface:
        def foo(self):                              # skipcq: PYL-R0201
            yield 1

    with pytest.raises(NotImplementedError):
        @implements(GenFooInterface)
        class FooImplementationFail:                # skipcq: PYL-W0612
            def foo(self):
                pass

    # must fail a generator which happens to be async
    with pytest.raises(NotImplementedError):
        @implements(GenFooInterface)
        class AsyncGenFooImplementationFail:        # skipcq: PYL-W0612
            async def foo(self):
                yield 1

    @implements(GenFooInterface)
    class GenFooImplementation:                     # skipcq: PYL-W0612
        def foo(self):                              # skipcq: PYL-R0201
            yield 1


def test_asyncgen_method():
    class AsyncGenFooInterface:
        async def foo(self):
            yield 1

    with pytest.raises(NotImplementedError):
        @implements(AsyncGenFooInterface)
        class AsyncFooImplementationFail:           # skipcq: PYL-W0612
            async def foo(self):
                pass

    with pytest.raises(NotImplementedError):
        @implements(AsyncGenFooInterface)
        class GenFooImplementationFail:             # skipcq: PYL-W0612
            def foo(self):                          # skipcq: PYL-R0201
                yield 1

    @implements(AsyncGenFooInterface)
    class AsyncGenFooImplementation:                # skipcq: PYL-W0612
        async def foo(self):
            yield 1


@py36
def test_new_style_descriptors():
    class IntField:
        def __get__(self, instance, owner):
            return instance.__dict__[self.name]

        def __set__(self, instance, value):
            if not isinstance(value, int):
                raise ValueError('expecting integer in {}'.format(self.name))
            instance.__dict__[self.name] = value

        def __set_name__(self, owner, name):
            self.name = name                            # skipcq: PYL-W0201

    class FooInterface(Interface):
        int_field = IntField()

    with pytest.raises(NotImplementedError):
        @implements(FooInterface)
        class FooImplementationFail:
            pass

    @implements(FooInterface)
    class FooImplementationPass:
        int_field = IntField()


@py36
def test_new_style_metaclasses():
    class Polygon:
        def __init_subclass__(cls, sides, **kwargs):
            cls.sides = sides
            if cls.sides < 3:
                raise ValueError('polygons need 3+ sides')

        @classmethod
        def interior_angles(cls):
            return (cls.sides - 2) * 180

    class PolygonInterface(Interface):
        def rotate(self):
            pass

    @implements(PolygonInterface)
    class Triangle(Polygon, sides=3):
        def rotate(self):
            pass


def test_descriptors_signature_getter():
    class FooInterface(Interface):
        @property
        def someprop(self) -> str:
            pass

    with pytest.raises(NotImplementedError):
        @implements(FooInterface)
        class FooImplementationFail:
            @property
            def someprop(self) -> int:
                pass


def test_descriptors_signature_setter():
    class FooInterface(Interface):
        @property
        def someprop(self):
            pass

        @someprop.setter
        def someprop(self, value: str) -> str:
            pass

    with pytest.raises(NotImplementedError):
        @implements(FooInterface)
        class FooImplementationFail:
            @property
            def someprop(self):
                pass

            @someprop.setter
            def someprop(self, value: int) -> float:
                pass


def test_descriptors_signature_deleter():
    class FooInterface(Interface):
        @property
        def someprop(self):
            pass

        @someprop.deleter
        def someprop(self) -> str:
            pass

    with pytest.raises(NotImplementedError):
        @implements(FooInterface)
        class FooImplementationFail:
            @property
            def someprop(self):
                pass

            @someprop.deleter
            def someprop(self) -> int:
                pass


def test_get_mro():
    class RegularClass:
        pass

    mro = get_mro(RegularClass)
    assert object not in mro

    expected = RegularClass.mro()[:-1]
    assert mro == expected


def test_class_hierarchy_overlap_of_common_class():
    class CommonClass:
        pass

    class FooInterface(CommonClass):
        def abc(self) -> str:
            pass

    with pytest.raises(ValueError):
        @implements(FooInterface)
        class FooImplemenation(CommonClass):
            def abc(self) -> str:
                pass


def test_implementation_inheriting_from_interface():
    class FooInterface:
        def abc(self) -> str:
            pass

    with pytest.raises(ValueError):
        @implements(FooInterface)
        class FooImplemenation(FooInterface):
            def abc(self) -> str:
                pass



"""Distutils setup file, used to install or test 'implements'."""
from __future__ import print_function

import textwrap
from setuptools import setup

with open('README.rst') as f:
    readme = f.read()


_DESCRIPTION = 'Pythonic interfaces using decorators'

_AUTHORS = 'Kamil Sindi <ksindi@ksindi.com>'

_INSTALL_REQUIRES = ['setuptools_scm>=1.15.0']
_SETUP_REQUIRES = ['setuptools>=18.0', 'pytest-runner',
                   'setuptools_scm>=1.15.0', 'sphinx_rtd_theme']
_TEST_REQUIRES = ['pytest', 'pytest-flake8']
_ALL_PACKAGES = []
_ALL_PACKAGES.extend(_SETUP_REQUIRES)
_ALL_PACKAGES.extend(_TEST_REQUIRES)

if __name__ == '__main__':
    setup(
        name='implements',
        description=_DESCRIPTION,
        long_description=readme,
        url='http://implements.readthedocs.io',
        use_scm_version=True,
        author=_AUTHORS,
        maintainer='Kamil Sindi',
        maintainer_email='ksindi@ksindi.com',
        install_requires=_INSTALL_REQUIRES,
        setup_requires=_SETUP_REQUIRES,
        tests_require=_TEST_REQUIRES,
        all_packages=_ALL_PACKAGES,
        zip_safe=False,
        include_package_data=True,
        py_modules=['implements'],
        classifiers=textwrap.dedent("""
            Development Status :: 5 - Production/Stable
            Intended Audience :: Developers
            License :: OSI Approved :: Apache Software License
            Natural Language :: English
            Programming Language :: Python :: 3
            Programming Language :: Python :: 3.6
            Programming Language :: Python :: 3.7
            Programming Language :: Python :: 3.8
            """).strip().splitlines(),
        keywords=['implements', 'interfaces'],
        license='Apache License, Version 2.0',
    )
